@@ -15,13 +15,20 @@ import hashlib
 import os
 import re
 import shlex
+import shutil
 import subprocess
 from contextlib import contextmanager
 
 from docutils import nodes
 from docutils.parsers.rst import directives
 from docutils.parsers.rst import Directive
+
+from sphinx import util
 from sphinx.errors import SphinxError
+from sphinx.util import (
+    i18n,
+    logging,
+)
 from sphinx.util.nodes import set_source_info
 from sphinx.util.osutil import (
     ensuredir,
@@ -33,20 +40,21 @@ try:
 except ImportError:
     Image = None
 
-try:
-    from sphinx.util.i18n import search_image_for_language
-except ImportError:  # Sphinx < 1.4
-    def search_image_for_language(filename, env):
-        return filename
 
-try:
-    from sphinx.util import logging
-    logger = logging.getLogger(__name__)
-    def _warn(self, msg):
-        logger.warning(msg)
-except (AttributeError, ImportError):  # Sphinx < 1.6
-    def _warn(self, msg):
-        self.builder.warn(msg)
+logger = logging.getLogger(__name__)
+
+
+if os.name == 'nt':
+    def rename(src, dst):
+        try:
+            os.rename(src, dst)
+        except OSError as err:
+            if err.errno != errno.EEXIST:
+                raise
+            os.unlink(dst)
+            os.rename(src, dst)
+else:
+    rename = os.rename
 
 
 class PlantUmlError(SphinxError):
@@ -105,7 +113,7 @@ class UmlDirective(Directive):
             return [warning('uml directive cannot have both content and '
                             'a filename argument', line=self.lineno)]
         if self.arguments:
-            fn = search_image_for_language(self.arguments[0], env)
+            fn = i18n.search_image_for_language(self.arguments[0], env)
             relfn, absfn = env.relfn2path(fn)
             env.note_dependency(relfn)
             try:
@@ -152,13 +160,17 @@ def _read_utf8(filename):
         fp.close()
 
 
-def generate_name(self, node, fileformat):
+def hash_plantuml_node(node):
     h = hashlib.sha1()
     # may include different file relative to doc
     h.update(node['incdir'].encode('utf-8'))
     h.update(b'\0')
     h.update(node['uml'].encode('utf-8'))
-    key = h.hexdigest()
+    return h.hexdigest()
+
+
+def generate_name(self, node, fileformat):
+    key = hash_plantuml_node(node)
     fname = 'plantuml-%s.%s' % (key, fileformat)
     imgpath = getattr(self.builder, 'imgpath', None)
     if imgpath:
@@ -203,30 +215,12 @@ def render_plantuml(self, node, fileformat):
     refname, outfname = generate_name(self, node, fileformat)
     if os.path.exists(outfname):
         return refname, outfname  # don't regenerate
-    absincdir = os.path.join(self.builder.srcdir, node['incdir'])
+
+    cachefname = self.builder.plantuml_builder.render(node, fileformat)
     ensuredir(os.path.dirname(outfname))
-    f = open(outfname, 'wb')
-    try:
-        try:
-            p = subprocess.Popen(generate_plantuml_args(self, node,
-                                                        fileformat),
-                                 stdout=f, stdin=subprocess.PIPE,
-                                 stderr=subprocess.PIPE,
-                                 cwd=absincdir)
-        except OSError as err:
-            if err.errno != ENOENT:
-                raise
-            raise PlantUmlError('plantuml command %r cannot be run'
-                                % self.builder.config.plantuml)
-        serr = p.communicate(node['uml'].encode('utf-8'))[1]
-        if p.returncode != 0:
-            if self.builder.config.plantuml_syntax_error_image:
-                _warn(self, 'error while running plantuml\n\n%s' % serr)
-            else:
-                raise PlantUmlError('error while running plantuml\n\n%s' % serr)
-        return refname, outfname
-    finally:
-        f.close()
+    # TODO: optionally do symlink/link
+    shutil.copyfile(cachefname, outfname)
+    return refname, outfname
 
 
 def render_plantuml_inline(self, node, fileformat):
@@ -248,6 +242,135 @@ def render_plantuml_inline(self, node, fileformat):
     return sout.decode('utf-8')
 
 
+class PlantumlBuilder(object):
+    def __init__(self, builder):
+        # for compatibility with existing functions which expect self.builder
+        # TODO: remove self.builder
+        self.builder = builder
+
+        self.batch_size = builder.config.plantuml_batch_size
+        self.cache_dir = os.path.join(builder.outdir,
+                                      builder.config.plantuml_cache_path)
+
+        self._base_cmdargs = _split_cmdargs(builder.config.plantuml)
+        self._base_cmdargs.extend(['-charset', 'utf-8'])
+
+        self.image_formats = []
+        if builder.format == 'html':
+            fmt = builder.config.plantuml_output_format
+            if fmt != 'none':
+                fileformats, _gettag = _lookup_html_format(fmt)
+                self.image_formats = list(fileformats)
+        elif builder.format == 'latex':
+            fmt = builder.config.plantuml_latex_output_format
+            if fmt != 'none':
+                fileformat, _postproc = _lookup_latex_format(fmt)
+                self.image_formats = [fmt]
+
+        self._known_keys = set()
+        self._pending_keys = []
+
+    def collect_nodes(self, doctree):
+        for node in doctree.traverse(plantuml):
+            key = hash_plantuml_node(node)
+            if key in self._known_keys:
+                continue
+            self._known_keys.add(key)
+
+            doc = node['uml'].encode('utf-8')
+            if b'!include' in doc or b'%filename' in doc:
+                # Heuristic to work around the path/filename issue. There's no
+                # easy way to specify the cwd of the doc without using -pipe.
+                continue
+
+            outdir = os.path.join(self.cache_dir, key[:2])
+            outfbase = os.path.join(outdir, key)
+            if all(os.path.exists('%s.%s' % (outfbase, sfx))
+                   for sfx in ['puml'] + self.image_formats):
+                continue
+
+            ensuredir(outdir)
+            with open(outfbase + '.puml', 'wb') as f:
+                # @startuml/enduml is mandatory in non-pipe mode. For backward
+                # compatibility, we do wrap the document only if @startXYZ is
+                # not specified.
+                started = doc.lstrip().startswith(b'@start')
+                if not started:
+                    f.write(b'@startuml\n')
+                f.write(doc)
+                if not started:
+                    f.write(b'\n@enduml\n')
+
+            self._pending_keys.append(key)
+
+    def render_batches(self):
+        pending_keys = sorted(self._pending_keys)
+        for fileformat in self.image_formats:
+            for i in range(0, len(pending_keys), self.batch_size):
+                keys = pending_keys[i:i + self.batch_size]
+                with util.progress_message(
+                        'rendering plantuml diagrams [%d..%d/%d]'
+                        % (i, i + len(keys), len(pending_keys))):
+                    self._render_files(keys, fileformat)
+
+        del self._pending_keys[:]
+
+    def _render_files(self, keys, fileformat):
+        cmdargs = self._base_cmdargs[:]
+        cmdargs.extend(_ARGS_BY_FILEFORMAT[fileformat])
+        cmdargs.extend(os.path.join(k[:2], '%s.puml' % k) for k in keys)
+        try:
+            p = subprocess.Popen(cmdargs, stderr=subprocess.PIPE,
+                                 cwd=self.cache_dir)
+        except OSError as err:
+            if err.errno != ENOENT:
+                raise
+            raise PlantUmlError('plantuml command %r cannot be run'
+                                % self.builder.config.plantuml)
+        serr = p.communicate()[1]
+        if p.returncode != 0:
+            if self.builder.config.plantuml_syntax_error_image:
+                logger.warning('error while running plantuml\n\n%s' % serr)
+            else:
+                raise PlantUmlError('error while running plantuml\n\n%s' % serr)
+
+    def render(self, node, fileformat):
+        key = hash_plantuml_node(node)
+        outdir = os.path.join(self.cache_dir, key[:2])
+        outfname = os.path.join(outdir, '%s.%s' % (key, fileformat))
+        if os.path.exists(outfname):
+            return outfname
+
+        ensuredir(outdir)
+        absincdir = os.path.join(self.builder.srcdir, node['incdir'])
+        with open(outfname + '.new', 'wb') as f:
+            try:
+                p = subprocess.Popen(generate_plantuml_args(self, node,
+                                                            fileformat),
+                                     stdout=f, stdin=subprocess.PIPE,
+                                     stderr=subprocess.PIPE,
+                                     cwd=absincdir)
+            except OSError as err:
+                if err.errno != ENOENT:
+                    raise
+                raise PlantUmlError('plantuml command %r cannot be run'
+                                    % self.builder.config.plantuml)
+            serr = p.communicate(node['uml'].encode('utf-8'))[1]
+            if p.returncode != 0:
+                if self.builder.config.plantuml_syntax_error_image:
+                    logger.warning('error while running plantuml\n\n%s' % serr)
+                else:
+                    raise PlantUmlError('error while running plantuml\n\n%s'
+                                        % serr)
+
+        rename(outfname + '.new', outfname)
+        return outfname
+
+
+def _render_batches_on_vist(self):
+    self.builder.plantuml_builder.render_batches()
+
+
 def _get_png_tag(self, fnames, node):
     refname, outfname = fnames['png']
     alt = node.get('alt', node['uml'])
@@ -256,9 +379,9 @@ def _get_png_tag(self, fnames, node):
     # process images prior to html_vist.
     scale_attrs = [k for k in ('scale', 'width', 'height') if k in node]
     if scale_attrs and Image is None:
-        _warn(self, ('plantuml: unsupported scaling attributes: %s '
-                     '(install PIL or Pillow)'
-                     % ', '.join(scale_attrs)))
+        logger.warning(('plantuml: unsupported scaling attributes: %s '
+                        '(install PIL or Pillow)'
+                        % ', '.join(scale_attrs)))
     if not scale_attrs or Image is None:
         return ('<img src="%s" alt="%s"/>\n'
                 % (self.encode(refname), self.encode(alt)))
@@ -289,7 +412,7 @@ def _get_png_tag(self, fnames, node):
             styles.extend('%s: %s%s' % (a, w * scale / 100, 'px')
                           for a, w in zip(['width', 'height'], im.size))
         except (IOError, OSError) as err:
-            _warn(self, 'plantuml: failed to get image size: %s' % err)
+            logger.warning('plantuml: failed to get image size: %s' % err)
 
     return ('<a href="%s"><img src="%s" alt="%s" style="%s"/>'
             '</a>\n'
@@ -352,27 +475,29 @@ _KNOWN_HTML_FORMATS = {
 }
 
 
+def _lookup_html_format(fmt):
+    try:
+        return _KNOWN_HTML_FORMATS[fmt]
+    except KeyError:
+        raise PlantUmlError(
+            'plantuml_output_format must be one of %s, but is %r'
+            % (', '.join(map(repr, _KNOWN_HTML_FORMATS)), fmt))
+
+
 @contextmanager
 def _prepare_html_render(self, fmt):
     if fmt == 'none':
         raise nodes.SkipNode
 
     try:
-        try:
-            fileformats, gettag = _KNOWN_HTML_FORMATS[fmt]
-        except KeyError:
-            raise PlantUmlError(
-                'plantuml_output_format must be one of %s, but is %r'
-                % (', '.join(map(repr, _KNOWN_HTML_FORMATS)), fmt))
-
-        yield fileformats, gettag
-
+        yield _lookup_html_format(fmt)
     except PlantUmlError as err:
-        _warn(self, str(err))
+        logger.warning(str(err))
         raise nodes.SkipNode
 
 
 def html_visit_plantuml(self, node):
+    _render_batches_on_vist(self)
     if 'html_format' in node:
         fmt = node['html_format']
     else:
@@ -420,7 +545,17 @@ _KNOWN_LATEX_FORMATS = {
 }
 
 
+def _lookup_latex_format(fmt):
+    try:
+        return _KNOWN_LATEX_FORMATS[fmt]
+    except KeyError:
+        raise PlantUmlError(
+            'plantuml_latex_output_format must be one of %s, but is %r'
+            % (', '.join(map(repr, _KNOWN_LATEX_FORMATS)), fmt))
+
+
 def latex_visit_plantuml(self, node):
+    _render_batches_on_vist(self)
     if 'latex_format' in node:
         fmt = node['latex_format']
     else:
@@ -428,16 +563,11 @@ def latex_visit_plantuml(self, node):
     if fmt == 'none':
         raise nodes.SkipNode
     try:
-        try:
-            fileformat, postproc = _KNOWN_LATEX_FORMATS[fmt]
-        except KeyError:
-            raise PlantUmlError(
-                'plantuml_latex_output_format must be one of %s, but is %r'
-                % (', '.join(map(repr, _KNOWN_LATEX_FORMATS)), fmt))
+        fileformat, postproc = _lookup_latex_format(fmt)
         refname, outfname = render_plantuml(self, node, fileformat)
         refname, outfname = postproc(self, refname, outfname)
     except PlantUmlError as err:
-        _warn(self, str(err))
+        logger.warning(str(err))
         raise nodes.SkipNode
 
     # put node representing rendered image
@@ -453,6 +583,7 @@ def latex_depart_plantuml(self, node):
 
 
 def confluence_visit_plantuml(self, node):
+    _render_batches_on_vist(self)
     fmt = self.builder.config.plantuml_output_format
     with _prepare_html_render(self, fmt) as (fileformats, _):
         _, outfname = render_plantuml(self, node, fileformats[0])
@@ -477,10 +608,11 @@ def confluence_visit_plantuml(self, node):
 
 
 def text_visit_plantuml(self, node):
+    _render_batches_on_vist(self)
     try:
         text = render_plantuml_inline(self, node, 'txt')
     except PlantUmlError as err:
-        _warn(self, str(err))
+        logger.warning(str(err))
         text = node['uml']  # fall back to uml text, which is still readable
 
     self.new_state()
@@ -490,18 +622,19 @@ def text_visit_plantuml(self, node):
 
 
 def pdf_visit_plantuml(self, node):
+    _render_batches_on_vist(self)
     try:
         refname, outfname = render_plantuml(self, node, 'eps')
         refname, outfname = _convert_eps_to_pdf(self, refname, outfname)
     except PlantUmlError as err:
-        _warn(self, str(err))
+        logger.warning(str(err))
         raise nodes.SkipNode
     rep = nodes.image(uri=outfname, alt=node.get('alt', node['uml']))
     node.parent.replace(node, rep)
 
 
 def unsupported_visit_plantuml(self, node):
-    _warn(self, 'plantuml: unsupported output format (node skipped)')
+    logger.warning('plantuml: unsupported output format (node skipped)')
     raise nodes.SkipNode
 
 
@@ -515,6 +648,25 @@ _NODE_VISITORS = {
 }
 
 
+def _on_builder_inited(app):
+    app.builder.plantuml_builder = PlantumlBuilder(app.builder)
+
+
+def _on_doctree_read(app, doctree):
+    # Collect as many static nodes as possible prior to start building.
+    if app.builder.plantuml_builder.batch_size > 1:
+        app.builder.plantuml_builder.collect_nodes(doctree)
+
+
+def _on_doctree_resolved(app, doctree, docname):
+    # Dynamically generated nodes will be collected here, which will be
+    # batched at node visitor. Since 'doctree-resolved' and node visits
+    # can be intermixed, there's no way to batch rendering of dynamic nodes
+    # at once.
+    if app.builder.plantuml_builder.batch_size > 1:
+        app.builder.plantuml_builder.collect_nodes(doctree)
+
+
 def setup(app):
     app.add_node(plantuml, **_NODE_VISITORS)
     app.add_directive('uml', UmlDirective)
@@ -528,6 +680,11 @@ def setup(app):
     app.add_config_value('plantuml_epstopdf', 'epstopdf', '')
     app.add_config_value('plantuml_latex_output_format', 'png', '')
     app.add_config_value('plantuml_syntax_error_image', False, '')
+    app.add_config_value('plantuml_cache_path', '_plantuml', '')
+    app.add_config_value('plantuml_batch_size', 1, '')
+    app.connect('builder-inited', _on_builder_inited)
+    app.connect('doctree-read', _on_doctree_read)
+    app.connect('doctree-resolved', _on_doctree_resolved)
 
     # imitate what app.add_node() does
     if 'rst2pdf.pdfbuilder' in app.config.extensions:
